from datetime import date, datetime
from functools import lru_cache
from logging import Logger
from os import PathLike
from pathlib import Path
import sqlite3
from sqlite3 import Connection, Cursor
from typing import Any, Dict, List, Mapping, Sequence, Union

from pyproj import CRS
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry

from tablecrow.tables.base import DatabaseTable, parse_record_values

SSH_DEFAULT_PORT = 22

GEOMETRY_TYPES = [
    'Point',
    'LineString',
    'Polygon',
    'MultiPoint',
    'MultiLineString',
    'MultiPolygon',
]


class SQLiteTable(DatabaseTable):
    FIELD_TYPES = {
        'NoneType': 'NULL',
        'bool': 'BOOLEAN',
        'float': 'REAL',
        'int': 'INTEGER',
        'str': 'TEXT',
        'date': 'DATE',
        'datetime': 'DATETIME',
        'bytes': 'BLOB',
        **{geometry_type: geometry_type.upper() for geometry_type in GEOMETRY_TYPES},
    }
    DEFAULT_PORT = None

    def __init__(
        self,
        path: PathLike,
        table_name: str,
        fields: Dict[str, type] = None,
        primary_key: Union[str, List[str]] = None,
        crs: CRS = None,
        logger: Logger = None,
    ):
        if '://' not in str(path):
            path = str(Path(path).expanduser().resolve())

        super().__init__(
            resource=path,
            table_name=table_name,
            database=None,
            fields=fields,
            primary_key=primary_key,
            crs=crs,
            logger=logger,
        )

        if not self.connected:
            raise ConnectionError(f'no connection to {self.path}/{self.name}')

        if self.fields is None:
            with self.connection:
                cursor = self.connection.cursor()
                self._DatabaseTable__fields = database_table_fields(cursor, self.name)

            if self.primary_key is None:
                self._DatabaseTable__primary_key = list(self.fields)[0]

        if len(self.geometry_fields) > 0:
            self.connection.enable_load_extension(True)

            try:
                self.connection.execute('SELECT load_extension("mod_spatialite")')
            except:
                import platform

                system_platform = platform.system()
                if system_platform == 'Windows':
                    message = 'download the module from `http://www.gaia-gis.it/gaia-sins/windows-bin-amd64/spatialite-loadable-modules-5.0.0-win-amd64.7z` and place `mod_spatialite.dll` / `mod_spatialite.so` in your `PATH`'
                elif system_platform in ['Linux', 'Darwin']:
                    message = (
                        'run the following: \n'
                        'sudo apt install libsqlite3-mod-spatialite \n'
                        'ln -sf /usr/lib/x86_64-linux-gnu/mod_spatialite.so /usr/lib/x86_64-linux-gnu/mod_spatialite'
                    )
                raise EnvironmentError(f'SpatiaLite module was not found; {message}')

        with self.connection:
            cursor = self.connection.cursor()
            if database_has_table(cursor, self.name):
                remote_fields = self.remote_fields
                if list(remote_fields) != list(self.fields):
                    self.logger.warning(
                        f'schema of existing table "{self.database}/{self.name}" differs from given fields'
                    )

                    remote_fields_not_in_local_table = {
                        field: value
                        for field, value in remote_fields.items()
                        if field not in self.fields
                    }
                    if len(remote_fields_not_in_local_table) > 0:
                        self.logger.warning(
                            f'remote table has {len(remote_fields_not_in_local_table)} fields not in local table: {list(remote_fields_not_in_local_table)}'
                        )
                        self.logger.warning(
                            f'adding {len(remote_fields_not_in_local_table)} fields to local table: {list(remote_fields_not_in_local_table)}'
                        )

                        self._DatabaseTable__fields.update(remote_fields_not_in_local_table)
                        self._DatabaseTable__fields = {
                            field: self._DatabaseTable__fields[field]
                            for field in remote_fields
                        }

                    local_fields_not_in_remote_table = {
                        field: value
                        for field, value in self.fields.items()
                        if field not in remote_fields
                    }
                    if len(local_fields_not_in_remote_table) > 0:
                        self.logger.warning(
                            f'local table has {len(local_fields_not_in_remote_table)} fields not in remote table: {list(local_fields_not_in_remote_table)}'
                        )
                        self.logger.warning(
                            f'adding {len(local_fields_not_in_remote_table)} fields to remote table: {list(local_fields_not_in_remote_table)}'
                        )

                    if list(remote_fields) != list(self.fields):
                        self.logger.warning(
                            f'altering schema of "{self.database}/{self.name}"'
                        )
                        self.logger.debug(self.remote_fields)
                        self.logger.debug(self.fields)

                        copy_table_name = f'old_{self.name}'

                        if database_has_table(cursor, copy_table_name):
                            cursor.execute(f'DROP TABLE {copy_table_name};')

                        cursor.execute(f'ALTER TABLE {self.name} RENAME TO {copy_table_name};')

                        cursor.execute(f'CREATE TABLE {self.name} ({self.schema});')
                        cursor.execute(f'PRAGMA table_info({copy_table_name})')
                        copy_table_fields = [record[1] for record in cursor.fetchall()]

                        cursor.execute(
                            f'INSERT INTO {self.name} ({", ".join(copy_table_fields)}) SELECT * FROM {copy_table_name};'
                        )

                        cursor.execute(f'DROP TABLE {copy_table_name};')
            else:
                self.logger.debug(f'creating remote table "{self.database}/{self.name}"')
                cursor.execute(f'CREATE TABLE {self.name} ({self.schema});')

    @property
    @lru_cache(maxsize=1)
    def path(self) -> Path:
        return Path(self.resource)

    @property
    @lru_cache(maxsize=1)
    def connection(self) -> Connection:
        return sqlite3.connect(database=self.path)

    @property
    def database(self) -> str:
        return str(self.path)

    @property
    def exists(self) -> bool:
        if self.connected:
            with self.connection:
                cursor = self.connection.cursor()
                return database_has_table(cursor, self.name)

    @property
    def schema(self) -> str:
        """ SQLite schema string """

        schema = []
        for field, field_type in self.fields.items():
            if (
                isinstance(field_type, Sequence)
                and not isinstance(field_type, str)
                or isinstance(field_type, Mapping)
            ):
                raise sqlite3.OperationalError('SQLite does not support arrays or mappings')

            try:
                field_type = self.FIELD_TYPES[field_type.__name__]
            except KeyError:
                raise TypeError(f'SQLite does not support type "{field_type}"')

            schema.append(f'{field} {field_type}')

        schema.append(f'PRIMARY KEY({", ".join(self.primary_key)})')

        return ', '.join(schema)

    @property
    def remote_fields(self) -> Dict[str, type]:
        if not self.connected:
            raise ConnectionError(f'no connection to {self.database}/{self.name}')

        geometry_fields = [geometry_type.lower() for geometry_type in GEOMETRY_TYPES]

        with self.connection:
            cursor = self.connection.cursor()
            if database_has_table(cursor, self.name):
                fields = database_table_fields(cursor, self.name)

                for field, field_type in fields.items():
                    field_type = field_type.lower()
                    if field_type in geometry_fields:
                        if field in self.fields:
                            fields[field] = self.fields[field]
                            continue

                    for python_type, sqlite_type in self.FIELD_TYPES.items():
                        if sqlite_type.lower() == field_type:
                            if field_type in geometry_fields:
                                if python_type not in globals():
                                    exec(f'from shapely.geometry import {python_type}')
                            field_type = eval(python_type)
                            break
                    else:
                        for python_type, sqlite_type in self.FIELD_TYPES.items():
                            if python_type.lower() in field_type:
                                field_type = eval(python_type)
                                break
                        else:
                            field_type = str

                    fields[field] = field_type
            else:
                fields = None

            return fields

    @property
    def connected(self) -> bool:
        connected = False
        if self.path.exists():
            with self.connection:
                try:
                    cursor = self.connection.cursor()
                    cursor.execute('SELECT 1;')
                    cursor.fetchone()
                    connected = True
                except:
                    connected = False
        return connected

    def records_where(
        self, where: Union[Mapping[str, Any], str, List[str]]
    ) -> List[Dict[str, Any]]:
        if not self.connected:
            raise ConnectionError(f'no connection to {self.database}/{self.name}')

        where_clause, where_values = self.__where_clause(where)

        with self.connection:
            cursor = self.connection.cursor()
            if where_clause is None:
                cursor.execute(f'SELECT {", ".join(self.fields.keys())} FROM {self.name}')
            else:
                try:
                    if where_values is not None:
                        cursor.execute(
                            f'SELECT * FROM {self.name} WHERE {where_clause}', where_values
                        )
                    else:
                        cursor.execute(f'SELECT * FROM {self.name} WHERE {where_clause}')
                except sqlite3.OperationalError:
                    raise
            matching_records = cursor.fetchall()

        matching_records = [
            parse_record_values(dict(zip(self.fields.keys(), record)), self.fields)
            for record in matching_records
        ]

        return matching_records

    def records_intersecting(
        self, geometry: BaseGeometry, crs: CRS = None, geometry_fields: List[str] = None
    ) -> List[Dict[str, Any]]:
        if crs is None:
            crs = self.crs

        if crs.to_epsg() is None:
            raise NotImplementedError(f'no EPSG code found for CRS "{crs}"')

        if geometry_fields is None or len(geometry_fields) == 0:
            geometry_fields = list(self.geometry_fields)

        where_clause = []
        where_values = []
        for field in geometry_fields:
            where_values.extend([geometry.wkt, crs.to_epsg()])
            geometry_string = f'GeomFromText(?, ?)'
            if crs != self.crs:
                geometry_string = f'Transform({geometry_string}, ?)'
                where_values.append(self.crs.to_epsg())
            where_clause.append(f'Intersects({field}, {geometry_string})')
        where_clause = ' OR '.join(where_clause)

        non_geometry_fields = {
            field: field_type
            for field, field_type in self.fields.items()
            if field_type.__name__ not in GEOMETRY_TYPES
        }

        with self.connection:
            cursor = self.connection.cursor()
            cursor.execute(
                f'SELECT {", ".join(non_geometry_fields)} '
                f'FROM {self.name} WHERE {where_clause};',
                where_values,
            )
            non_geometry_records = cursor.fetchall()
            non_geometry_records = [
                parse_record_values(
                    dict(zip(non_geometry_fields.keys(), record)), non_geometry_fields
                )
                for record in non_geometry_records
            ]

            geometry_field_string = ', '.join(
                f'asbinary({geometry_field})' for geometry_field in self.geometry_fields
            )
            cursor.execute(
                f'SELECT {geometry_field_string} ' f'FROM {self.name} WHERE {where_clause};',
                where_values,
            )
            geometry_records = cursor.fetchall()
            geometry_records = [
                parse_record_values(
                    dict(zip(self.geometry_fields.keys(), record)), self.geometry_fields
                )
                for record in geometry_records
            ]

        records = [
            {**non_geometry_records[index], **geometry_records[index]}
            for index in range(len(non_geometry_records))
        ]
        return [{field: record[field] for field in self.fields} for record in records]

    def insert(self, records: List[Dict[str, Any]]):
        if isinstance(records, dict):
            records = [records]

        if not all(field in record for field in self.primary_key for record in records):
            raise KeyError(
                f'one or more records does not contain primary key(s) "{self.primary_key}"'
            )

        if not self.connected:
            raise ConnectionError(f'no connection to {self.database}/{self.name}')

        with self.connection:
            cursor = self.connection.cursor()
            for record in records:
                if len(self.primary_key) == 1:
                    primary_key_string = self.primary_key[0]
                    primary_key_value = record[self.primary_key[0]]
                else:
                    primary_key_string = f'({", ".join(self.primary_key)})'
                    primary_key_value = tuple(
                        record[primary_key] for primary_key in self.primary_key
                    )

                record_fields_not_in_local_table = [
                    field for field in record if field not in self.fields
                ]
                if len(record_fields_not_in_local_table) > 0:
                    self.logger.warning(
                        f'record has {len(record_fields_not_in_local_table)} fields not in the local table'
                        f' that will not be inserted: {record_fields_not_in_local_table}'
                    )

                local_fields_in_record = [field for field in self.fields if field in record]
                geometry_fields = [field for field in self.geometry_fields if field in record]

                columns = [
                    field for field in local_fields_in_record if field not in geometry_fields
                ]
                values = [
                    record[field]
                    for field in local_fields_in_record
                    if field not in geometry_fields
                ]

                if primary_key_value in self:
                    record_without_primary_key = {
                        column: value
                        for column, value in zip(columns, values)
                        if column not in self.primary_key
                    }
                    if not isinstance(primary_key_value, Sequence) or isinstance(
                        primary_key_value, str
                    ):
                        primary_key_value = [primary_key_value]
                    if len(record_without_primary_key) > 0:
                        cursor.execute(
                            f'UPDATE {self.name} '
                            f'SET ({", ".join(record_without_primary_key.keys())}) = ({", ".join("?" for _ in record_without_primary_key)})'
                            f' WHERE {primary_key_string} = ({", ".join("?" for _ in primary_key_value)});',
                            [*record_without_primary_key.values(), *primary_key_value,],
                        )
                else:
                    cursor.execute(
                        f'INSERT INTO {self.name} ({", ".join(columns)}) VALUES ({", ".join("?" for _ in values)});',
                        values,
                    )

                if len(geometry_fields) > 0:
                    geometries = {
                        field: record[field]
                        for field in geometry_fields
                        if record[field] is not None
                    }

                    for field, geometry in geometries.items():
                        cursor.execute(
                            f'UPDATE {self.name} SET {field} = GeomFromText(?, ?) '
                            f'WHERE {primary_key_string} = ?;',
                            [geometry.wkt, self.crs.to_epsg(), primary_key_value],
                        )

    def delete_where(self, where: Union[Mapping[str, Any], str, List[str]]):
        if not self.connected:
            raise ConnectionError(f'no connection to {self.database}/{self.name}')

        where_clause, where_values = self.__where_clause(where)

        with self.connection:
            cursor = self.connection.cursor()
            if where_clause is None:
                cursor.execute(f'TRUNCATE {self.name};')
            else:
                try:
                    cursor.execute(
                        f'DELETE FROM {self.name} WHERE {where_clause};', where_values
                    )
                except sqlite3.OperationalError as error:
                    raise SyntaxError(f'invalid SQL syntax - {error}')
                except sqlite3.DatabaseError as error:
                    raise KeyError(error)

    def __len__(self) -> int:
        with self.connection:
            cursor = self.connection.cursor()
            cursor.execute(f'SELECT COUNT(*) FROM {self.name};')
            return cursor.fetchone()[0]

    def delete_table(self):
        with self.connection:
            cursor = self.connection.cursor()
            cursor.execute(f'DROP TABLE {self.name};')

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}({repr(self.database)}, {repr(self.name)}, '
            f'{repr(self.fields)}, {repr(self.primary_key)}, {repr(self.crs.to_epsg()) if self.crs is not None else None})'
        )

    def __where_clause(self, where: Dict[str, Union[Any, list]]) -> (str, List[Any]):
        if (
            where is not None
            and not isinstance(where, Sequence)
            and not isinstance(where, dict)
        ):
            raise NotImplementedError(f'unsupported query type "{type(where)}"')

        if where is None or len(where) == 0:
            where_clause = None
            where_values = None
        else:
            fields = None
            where_values = []
            if isinstance(where, str):
                where_clause = where
            elif isinstance(where, dict):
                where_clause = []
                for field, value in where.items():
                    field_type = self.fields[field]
                    if isinstance(value, BaseGeometry) or isinstance(
                        value, BaseMultipartGeometry
                    ):
                        where_clause.append(f'{field} = GeomFromText(?, ?)')
                        where_values.extend([value.wkt, self.crs.to_epsg()])
                    else:
                        if isinstance(field_type, list):
                            if not isinstance(value, Sequence) or isinstance(value, str):
                                statement = f'? = ANY({field})'
                            else:
                                if fields is None:
                                    with self.connection:
                                        cursor = self.connection.cursor()
                                        fields = database_table_fields(cursor, self.name)
                                field_type = fields[field]
                                dimensions = field_type.count('_')
                                field_type = field_type.strip('_')
                                statement = f'{field} = ?::{field_type}{"[]" * dimensions}'
                        elif value is None:
                            statement = f'{field} IS ?'
                        elif isinstance(value, Sequence) and not isinstance(value, str):
                            statement = f'{field} IN ({", ".join("?" for _ in value)})'
                        elif isinstance(value, str) and '%' in value:
                            statement = f'UPPER({field}) LIKE ?'
                            value = value.upper()
                        else:
                            if isinstance(value, datetime) or isinstance(value, date):
                                value = f'{value:%Y-%m-%d %H:%M:%S}'
                            statement = f'{field} = ?'
                        if isinstance(value, Sequence) and not isinstance(value, str):
                            where_values.extend(value)
                        else:
                            where_values.append(value)
                        where_clause.append(statement)
                where_clause = ' AND '.join(where_clause)
            else:
                where_clause = ' AND '.join(where)

            if len(where_values) == 0:
                where_values = None

        return where_clause, where_values


def database_tables(cursor: Cursor) -> List[str]:
    """
    list of tables within the given SQLite database

    :param cursor: sqlite3 cursor
    :return: list of table names
    """

    cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table';")
    return [record[0] for record in cursor.fetchall()]


def database_has_table(cursor: Cursor, table: str) -> bool:
    """
    whether the given table exists within the given SQLite database

    :param cursor: sqlite3 cursor
    :param table: name of table
    :return: whether table exists
    """

    return table in database_tables(cursor)


def database_table_fields(cursor: Cursor, table: str) -> Dict[str, str]:
    """
    field names and data types of the given table, within the given SQLite database

    :param cursor: sqlite3 cursor
    :param table: name of table
    :return: mapping of column names to the SQLite data type
    """

    cursor.execute(f'PRAGMA table_info({table});')
    return {record[1]: record[2] for record in cursor.fetchall()}



from os import PathLike
from pathlib import Path
import sqlite3
from typing import List, Union

import psycopg2

from tablecrow.tables.base import DatabaseTable
from tablecrow.tables.postgres import PostGresTable
from tablecrow.tables.postgres import database_tables as postgres_database_tables
from tablecrow.tables.sqlite import SQLiteTable
from tablecrow.tables.sqlite import database_tables as sqlite_database_tables
from tablecrow.utilities import parse_hostname

DATABASE_FUNCTIONS = {
    'PostGres': {
        'connect': psycopg2.connect,
        'table_names': postgres_database_tables,
        'table': PostGresTable,
    },
    'SQLite': {
        'connect': sqlite3.connect,
        'table_names': sqlite_database_tables,
        'table': SQLiteTable,
    },
}


def connect(
    resource: Union[str, PathLike], table_names: List[str] = None, **kwargs
) -> List[DatabaseTable]:
    """
    the main connection function for connecting to an arbitrary database table

    `tablecrow.connect()` will return TableCrow table(s) based on the given connection parameters

    :param resource: location of database
    :param table_names: names of tables to connect to
    :return: list of tables
    """

    if table_names is None:
        table_names = []

    database_type = None
    connection = None
    messages = []

    credentials = parse_hostname(resource)

    database_types = list(DATABASE_FUNCTIONS)

    path = Path(credentials['hostname']).expanduser()
    if path.exists() or path.anchor != '':
        resource = Path(resource).resolve().absolute()
        database_types = ['SQLite']
        credentials = {'database': credentials['hostname']}
    else:
        database_types.remove('SQLite')

    for credential in credentials:
        if credential in kwargs and kwargs[credential] is not None:
            credentials[credential] = kwargs[credential]

    for current_database_type in database_types:
        try:
            current_credentials = credentials.copy()
            if current_database_type == 'PostGres':
                current_credentials['host'] = current_credentials['hostname']
                current_credentials['user'] = current_credentials['username']
                del current_credentials['hostname'], current_credentials['username']
            connection = DATABASE_FUNCTIONS[current_database_type]['connect'](
                **current_credentials
            )
            database_type = current_database_type
            break
        except Exception as error:
            messages.append(f'{current_database_type} - {error.__class__.__name__}: {error}')

    if database_type is not None:
        functions = DATABASE_FUNCTIONS[database_type]
        with connection:
            cursor = connection.cursor()
            table_names.extend(functions['table_names'](cursor))
            if 'table_name' in kwargs:
                table_names.append(kwargs['table_name'])
                del kwargs['table_name']
            table_names = set(table_names)
            return [
                functions['table'](resource, table_name=table_name, **kwargs)
                for table_name in table_names
            ]
    else:
        message = ''.join(messages)
        raise ConnectionError(
            f'could not connect to "{resource}" with {[functions["connect"].__module__ + "." + functions["connect"].__name__ for functions in DATABASE_FUNCTIONS.values()]}'
            f'\n'
            f'{message}'
        )



from abc import ABC, abstractmethod
import logging
from logging import Logger
from pathlib import Path
import socket
from typing import Any, Dict, Generator, List, Mapping, Sequence, Union

from pyproj import CRS
from shapely.geometry import LinearRing, MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry, GEOMETRY_TYPES

from tablecrow.utilities import convert_value, get_logger, parse_hostname

DEFAULT_CRS = CRS.from_epsg(4326)


class TableNotFoundError(FileNotFoundError):
    pass


class DatabaseTable(ABC):
    DEFAULT_PORT = NotImplementedError
    FIELD_TYPES: Dict[str, str] = NotImplementedError

    def __init__(
        self,
        resource: str,
        table_name: str,
        database: str = None,
        fields: Dict[str, type] = None,
        primary_key: Union[str, List[str]] = None,
        crs: CRS = None,
        username: str = None,
        password: str = None,
        users: List[str] = None,
        logger: Logger = None,
    ):
        """
        abstraction of a database table

        :param resource: URL of database server as `hostname:port`
        :param table_name: name of table in database
        :param database: name of database in server
        :param fields: dictionary of fields
        :param primary_key: primary key field(s)
        :param crs: coordinate reference system of table geometries
        :param username: username to connect ot database
        :param password: password to connect to database
        :param users: list of database users / roles to give access
        """

        self.__database = database
        self.__name = table_name
        self.__fields = fields

        if logger is None:
            logger = get_logger('dummy', console_level=logging.NOTSET)

        self.logger = logger

        if resource is not None:
            if Path(resource).exists():
                port = None
            else:
                credentials = parse_hostname(resource)
                resource = credentials['hostname']
                port = credentials['port']
                if port is None:
                    port = self.DEFAULT_PORT
                if username is None:
                    username = credentials['username']
                if password is None:
                    password = credentials['password']
        else:
            port = None

        self.__resource = resource
        self.__port = port

        if username is not None and ':' in username:
            username, password = username.split(':', 1)

        self.__username = username
        self.__password = password

        if primary_key is None:
            primary_key = [list(self.fields)[0]] if self.fields is not None else None
        elif not isinstance(primary_key, Sequence) or isinstance(primary_key, str):
            primary_key = [primary_key]

        self.__primary_key = primary_key

        if crs is not None:
            crs = parse_crs(crs)
        elif len(self.geometry_fields) > 0:
            crs = DEFAULT_CRS
            self.logger.warning(
                f'no CRS provided for geometry fields; defaulting to EPSG:{crs.to_epsg()}'
            )
        else:
            crs = None
        self.__crs = crs

        if users is None:
            users = []

        self.__users = users

        if self.fields is None and not self.exists:
            raise TableNotFoundError(
                f'fields must be specified when creating a table; table does not exist at "{self.database}:{self.name}"'
            )

    @property
    def resource(self) -> str:
        return self.__resource

    @property
    def port(self) -> int:
        return self.__port

    @property
    def database(self) -> str:
        return self.__database

    @property
    def name(self) -> str:
        return self.__name

    @property
    def fields(self) -> Dict[str, type]:
        if self.__fields is None:
            self.__fields = self.remote_fields
        return self.__fields

    @property
    def primary_key(self) -> List[str]:
        return self.__primary_key

    @property
    def crs(self) -> CRS:
        return self.__crs

    @property
    def username(self) -> str:
        return self.__username

    @property
    def users(self) -> List[str]:
        return self.__users

    @property
    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def schema(self) -> str:
        """ SQL schema string """
        raise NotImplementedError

    @property
    def geometry_fields(self) -> Dict[str, type]:
        """ local fields with geometry type """
        geometry_fields = {}
        if self.fields is not None:
            for field, field_type in self.fields.items():
                while isinstance(field_type, Sequence) and not isinstance(field_type, str):
                    if len(field_type) > 0:
                        field_type = field_type[0]
                    else:
                        field_type = list
                if isinstance(field_type, type) and field_type.__name__ in GEOMETRY_TYPES:
                    geometry_fields[field] = field_type
        return geometry_fields

    @property
    @abstractmethod
    def remote_fields(self) -> Dict[str, type]:
        """ fields at remote table """
        raise NotImplementedError

    @property
    def connected(self) -> bool:
        """ whether network connection exists to database server """
        try:
            socket.setdefaulttimeout(2)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect(
                (self.resource, self.port)
            )
            return True
        except socket.error:
            return False

    @property
    def records(self) -> List[Dict[str, Any]]:
        """ list of records in the table """
        return self.records_where(None)

    @abstractmethod
    def records_where(
        self, where: Union[Mapping[str, Any], str, List[str]]
    ) -> List[Dict[str, Any]]:
        """
        list of records in the table that match the query

        :param where: dictionary mapping keys to values, with which to match records
        :return: dictionaries of matching records
        """

        raise NotImplementedError

    @abstractmethod
    def insert(self, records: List[Dict[str, Any]]):
        """
        insert the list of records into the table

        :param records: dictionary records
        """

        raise NotImplementedError

    @abstractmethod
    def delete_where(self, where: Union[Mapping[str, Any], str, List[str]]):
        """
        delete records from the table matching the given query

        :param where: dictionary mapping keys to values, with which to match records
        """

        raise NotImplementedError

    def __getitem__(self, key: Any) -> Dict[str, Any]:
        """
        Return the record matching the given primary key value.

        :param key: value of primary key
        :return: dictionary record
        """

        if isinstance(key, dict):
            if not all(field in key for field in self.primary_key):
                raise ValueError(f'does not contain "{self.primary_key}"')
            where = key
        else:
            if isinstance(key, Generator):
                key = list(key)
            elif not isinstance(key, Sequence) or isinstance(key, str):
                key = [key]
            if len(key) != len(self.primary_key):
                raise ValueError(f'ambiguous value for primary key "{self.primary_key}"')
            where = {field: key[index] for index, field in enumerate(self.primary_key)}

        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.resource}:{self.port}/{self.database}/{self.name}'
            )

        try:
            records = self.records_where(where)

            if len(records) > 1:
                self.logger.warning(
                    f'found more than one record matching query {where}: {records}'
                )

            if len(records) > 0:
                return records[0]
            else:
                raise KeyError(f'no record with primary key "{key}"')
        except:
            raise KeyError(f'no record with primary key "{key}"')

    def __setitem__(self, key: Any, record: Dict[str, Any]):
        """
        Insert the given record into the table.

        :param key: value of primary key at which to insert record
        :param record: dictionary record
        """

        if isinstance(key, Generator):
            key = list(key)
        elif isinstance(key, dict):
            if not all(field in key for field in self.primary_key):
                raise KeyError(f'does not contain "{self.primary_key}"')
            key = [key[field] for field in self.primary_key]
        elif not isinstance(key, Sequence) or isinstance(key, str):
            key = [key]

        for key_index, primary_key in enumerate(self.primary_key):
            record[primary_key] = key[key_index]

        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.resource}:{self.port}/{self.database}/{self.name}'
            )

        self.insert([record])

    def __delitem__(self, key: Any):
        """
        Delete the record matching the given primary key value.

        :param key: value of primary key
        """

        if isinstance(key, dict):
            if not all(field in key for field in self.primary_key):
                raise ValueError(f'does not contain "{self.primary_key}"')
            where = key
        else:
            if isinstance(key, Generator):
                key = list(key)
            elif not isinstance(key, Sequence) or isinstance(key, str):
                key = [key]
            if len(key) != len(self.primary_key):
                raise ValueError(f'ambiguous value for primary key "{self.primary_key}"')
            where = {field: key[index] for index, field in enumerate(self.primary_key)}

        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.resource}:{self.port}/{self.database}/{self.name}'
            )

        try:
            self.delete_where(where)
        except:
            raise KeyError(f'no record with primary key "{key}"')

    def __len__(self) -> int:
        return len(self.records)

    def __contains__(self, key: Any) -> bool:
        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.resource}:{self.port}/{self.database}/{self.name}'
            )

        try:
            self[key]
            return True
        except KeyError:
            return False

    def __iter__(self) -> Generator:
        yield from self.records

    @abstractmethod
    def delete_table(self):
        raise NotImplementedError

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}({repr(self.database)}, {repr(self.name)}, {repr(self.fields)}, {repr(self.primary_key)}, '
            f'{repr(self.resource)}, {repr(self.username)}, {repr("*" * len(self.password))}, {repr(self.users)})'
        )


def random_open_tcp_port() -> int:
    """
    :return: get an open TCP port on the current host
    """

    open_socket = socket.socket()
    open_socket.bind(('', 0))
    return open_socket.getsockname()[1]


def crs_key(crs: CRS) -> str:
    """
    :param crs: coordinate reference system
    :return: CRS key
    """

    if not isinstance(crs, CRS):
        crs = parse_crs(crs)
    return crs.wkt.split('"')[1]


def is_compound_crs(crs: CRS) -> bool:
    """
    :param crs: coordinate reference system
    :return: whether the CRS is compound
    """

    if not isinstance(crs, CRS):
        crs = parse_crs(crs)
    return 'COMPD_CS' in crs.wkt or 'COMPOUNDCRS' in crs.wkt


def split_compound_crs(crs: CRS) -> List[CRS]:
    """
    split the given compound coordinate reference system into its constituent CRS parts

    :param crs: compound coordinate reference system
    :returns: list of CRS parts
    """

    if type(crs) is not CRS:
        crs = parse_crs(crs)

    if is_compound_crs(crs):
        working_string = crs.wkt

        # remove the compound CRS keyword and name from the string, along with the closing bracket
        working_string = working_string.split(',', 1)[-1][:-1]

        wkts = []
        while len(working_string) > 0:
            opening_brackets = 0
            closing_brackets = 0
            for index, character in enumerate(working_string):
                if character == '[':
                    opening_brackets += 1
                elif character == ']':
                    closing_brackets += 1

                if opening_brackets > 0 and opening_brackets == closing_brackets:
                    wkts.append(working_string[: index + 1])
                    working_string = working_string[index + 2 :]
                    break
            else:
                wkts.append(working_string)
                break

        return [CRS.from_string(wkt) for wkt in wkts]


def compound_crs(crs_list: List[CRS], key: str = None) -> CRS:
    """
    build a compound coordinate reference system from the provided list of constituent CRSs

    :param crs_list: list of coordinate reference systems
    :param key: name of CRS
    :returns: compound CRS
    """

    crs_list = [crs if type(crs) is CRS else parse_crs(crs) for crs in crs_list]

    if key is None:
        key = ' + '.join(crs_key(crs) for crs in crs_list)

    # TODO is keyword specced as COMPOUNDCRS?
    return CRS.from_string(f'COMPD_CS["{key}", {", ".join(crs.wkt for crs in crs_list)}]')


def parse_crs(crs: Union[str, int]) -> CRS:
    """
    parse a CRS object from the given well-known text or EPSG code

    :param crs: coordinate reference system; either well-known text or an EPSG code
    :returns: CRS object
    """

    if isinstance(crs, CRS):
        return crs
    elif (
        isinstance(crs, str)
        and '+' in crs
        and 'COMPD_CS' not in crs
        and 'COMPOUNDCRS' not in crs
    ):
        return compound_crs([parse_crs(crs_part.strip()) for crs_part in crs.split('+')])
    else:
        try:
            return CRS.from_epsg(int(crs))
        except ValueError:
            return CRS.from_string(str(crs))


def flatten_geometry(geometry: BaseGeometry) -> BaseGeometry:
    """
    :param geometry: Shapely geometry object
    :return: geometry with Z values removed
    """

    geometry_type = type(geometry)

    # strip 3rd dimension
    if 'POLYGON Z' in geometry.wkt:
        polygons = (
            [polygon for polygon in geometry] if geometry_type is MultiPolygon else [geometry]
        )
        for polygon_index, polygon in enumerate(polygons):
            exterior_2d = LinearRing([vertex[:2] for vertex in polygon.exterior.coords])
            interiors_2d = [
                LinearRing([vertex[:2] for vertex in interior.coords])
                for interior in polygon.interiors
            ]
            polygons[polygon_index] = Polygon(exterior_2d, interiors_2d)
        geometry = (
            MultiPolygon(polygons) if geometry_type is MultiPolygon else Polygon(polygons[0])
        )

    if not geometry.is_valid:
        geometry = geometry.buffer(0)
    return geometry


def parse_record_values(
    record: Dict[str, Any], field_types: Dict[str, type]
) -> Dict[str, Any]:
    """
    parse the values in the given record into their respective field types

    :param record: dictionary mapping fields to values
    :param field_types: dictionary mapping fields to types
    :return: record with values parsed into their respective types
    """

    for field, value in record.items():
        if field in field_types:
            field_type = field_types[field]
            record[field] = convert_value(value, field_type)
    return record



import configparser
from datetime import date, datetime, timedelta
from enum import Enum, EnumMeta
import json
import logging
from os import PathLike
from pathlib import Path
import re
import sys
from typing import Any, Collection, Dict, Iterable, Mapping, Union

from dateutil.parser import parse as parse_date
from pyproj import CRS
from shapely import wkb, wkt
from shapely.geometry import shape as shapely_shape
from shapely.geometry.base import GEOMETRY_TYPES


def read_configuration(filename: PathLike) -> Dict[str, str]:
    """
    read a configuration file with `configparser`

    :param filename: path to configuration
    :return: dictionary mapping of configuration entries
    """

    configuration_file = configparser.ConfigParser()
    configuration_file.read(filename)
    return {
        section_name: {key: value for key, value in section.items()}
        for section_name, section in configuration_file.items()
        if section_name.upper() != 'DEFAULT'
    }


def repository_root(path: PathLike = None) -> Path:
    """
    get the root directory of the current Git repository

    :param path: query path
    :return: repository root directory
    """

    if path is None:
        path = __file__
    if not isinstance(path, Path):
        path = Path(path)
    if path.is_file():
        path = path.parent
    if '.git' in (child.name for child in path.iterdir()) or path == path.parent:
        return path
    else:
        return repository_root(path.parent)


def split_hostname_port(hostname: str) -> (str, Union[str, None]):
    """
    split the given URL into host and port, assuming port is appended after a colon

    :param hostname: hostname string
    :return: hostname and port (if found, otherwise `None`)
    """

    port = None

    if ':' in hostname:
        parts = hostname.rsplit(':', 1)
        try:
            port = int(parts[-1])
            hostname = parts[0]
        except ValueError:
            pass

    return hostname, port


def parse_hostname(hostname: str) -> Dict[str, str]:
    """
    parse detailed connection information from the given URL

    :param hostname: URL with connection information
    :return: dictionary of hostname, port, username, and password
    """

    username = None
    password = None

    hostname, port = split_hostname_port(hostname)

    protocol_pattern = re.compile(r'^(?:http|ftp)s?://')
    result = re.search(protocol_pattern, hostname)
    protocol = result.group(0) if result is not None else ''
    hostname = re.sub(protocol_pattern, '', hostname)

    if '@' in hostname:
        username, hostname = hostname.split('@', 1)

    hostname = protocol + hostname

    if username is not None and ':' in username:
        username, password = username.split(':', 1)

    return {
        'hostname': hostname,
        'port': port,
        'username': username,
        'password': password,
    }


def guard_generic_alias(generic_alias) -> type:
    """
    convert an instance of a subscripted `typing._GenericAlias` to a subscripted type
    :param generic_alias: generic alias
    :return: type

    :example:
    >>> from typing import List
    >>> guard_generic_alias(List[str])
    [str]

    :example:
    >>> from typing import Dict
    >>> guard_generic_alias(Dict[str, float])
    {str: float}

    :example:
    >>> from typing import Dict, Tuple
    >>> guard_generic_alias({str: (Dict[int, str], str)})
    {str: ({int: str}, str)}

    """

    if hasattr(generic_alias, '__origin__'):
        type_class = generic_alias.__origin__
        if hasattr(generic_alias, '__args__'):
            members = generic_alias.__args__
            if issubclass(type_class, Mapping):
                members = [members]
        else:
            members = ()
    elif isinstance(generic_alias, Collection) and not isinstance(
        generic_alias, (EnumMeta, str)
    ):
        type_class = generic_alias.__class__
        if issubclass(type_class, Mapping):
            members = generic_alias.items()
        else:
            members = generic_alias
    else:
        return generic_alias

    members = [guard_generic_alias(member) for member in members]
    if type_class != generic_alias.__class__ or members != generic_alias:
        return type_class(members)
    else:
        return generic_alias


def convert_value(value: Any, to_type: type) -> Any:
    if isinstance(to_type, str):
        to_type = eval(to_type)

    if isinstance(value, Enum):
        value = value.name

    to_type = guard_generic_alias(to_type)

    if to_type is None:
        value = None
    elif isinstance(to_type, Collection):
        collection_type = type(to_type)
        if collection_type is not EnumMeta:
            if not issubclass(collection_type, Mapping):
                if value is not None:
                    to_type = list(to_type)
                    if not isinstance(value, Iterable) or isinstance(value, str):
                        value = [value]
                    if len(to_type) == 1:
                        to_type = [to_type[0] for _ in value]
                    elif len(to_type) == len(value):
                        to_type = to_type[: len(value)]
                    else:
                        raise ValueError(
                            f'unable to convert list of values of length {len(value)} '
                            f'to list of types of length {len(to_type)}: '
                            f'{value} -/> {to_type}'
                        )
                    value = collection_type(
                        convert_value(value[index], current_type)
                        for index, current_type in enumerate(to_type)
                    )
                else:
                    value = collection_type()
            elif isinstance(value, str):
                value = json.loads(value)
            elif isinstance(value, CRS):
                value = value.to_json_dict()
        elif value is not None:
            try:
                value = to_type[value]
            except (KeyError, ValueError):
                try:
                    value = to_type(value)
                except (KeyError, ValueError):
                    raise ValueError(
                        f'unrecognized entry "{value}"; must be one of {list(to_type)}'
                    )
    elif not isinstance(value, to_type) and value is not None:
        if issubclass(to_type, (datetime, date)):
            value = parse_date(value)
            if issubclass(to_type, date) and not issubclass(to_type, datetime):
                value = value.date()
        elif issubclass(to_type, timedelta):
            try:
                try:
                    time = datetime.strptime(value, '%H:%M:%S')
                    value = timedelta(
                        hours=time.hour, minutes=time.minute, seconds=time.second
                    )
                except:
                    parts = [float(part) for part in value.split(':')]
                    if len(parts) > 3:
                        days = parts.pop(0)
                    else:
                        days = 0
                    value = timedelta(
                        days=days, hours=parts[0], minutes=parts[1], seconds=parts[2]
                    )
            except:
                value = timedelta(seconds=float(value))
        elif to_type.__name__ in GEOMETRY_TYPES:
            try:
                value = wkb.loads(value, hex=True)
            except:
                try:
                    value = wkt.loads(value)
                except:
                    try:
                        value = wkb.loads(value)
                    except TypeError:
                        if isinstance(value, str):
                            value = eval(value)
                        try:
                            value = shapely_shape(value)
                        except:
                            value = to_type(value)
        elif issubclass(to_type, bool):
            try:
                value = eval(f'{value}')
            except:
                value = bool(value)

        if not isinstance(value, to_type):
            if isinstance(value, timedelta):
                if issubclass(to_type, str):
                    hours, remainder = divmod(value, timedelta(hours=1))
                    minutes, remainder = divmod(remainder, timedelta(minutes=1))
                    seconds = remainder / timedelta(seconds=1)
                    value = f'{hours:02}:{minutes:02}:{seconds:04.3}'
                else:
                    value /= timedelta(seconds=1)
            elif isinstance(value, CRS):
                if issubclass(to_type, str):
                    value = value.to_wkt()
                elif issubclass(to_type, dict):
                    value = value.to_json_dict()
                elif issubclass(to_type, int):
                    value = value.to_epsg()
            elif type(value).__name__ in GEOMETRY_TYPES and to_type.__name__ in GEOMETRY_TYPES:
                raise NotImplementedError('casting between geometric types not implemented')
            elif isinstance(value, (str, bytes)):
                try:
                    value = to_type.from_string(value)
                except:
                    value = to_type(value)
            else:
                value = to_type(value)

    return value


def get_logger(
    name: str,
    log_filename: PathLike = None,
    file_level: int = None,
    console_level: int = None,
    log_format: str = None,
) -> logging.Logger:
    if file_level is None:
        file_level = logging.DEBUG
    if console_level is None:
        console_level = logging.INFO
    logger = logging.getLogger(name)

    # check if logger is already configured
    if logger.level == logging.NOTSET and len(logger.handlers) == 0:
        # check if logger has a parent
        if '.' in name:
            logger.parent = get_logger(name.rsplit('.', 1)[0])
        else:
            # otherwise create a new split-console logger
            logger.setLevel(logging.DEBUG)
            if console_level != logging.NOTSET:
                if console_level <= logging.INFO:

                    class LoggingOutputFilter(logging.Filter):
                        def filter(self, rec):
                            return rec.levelno in (logging.DEBUG, logging.INFO)

                    console_output = logging.StreamHandler(sys.stdout)
                    console_output.setLevel(console_level)
                    console_output.addFilter(LoggingOutputFilter())
                    logger.addHandler(console_output)

                console_errors = logging.StreamHandler(sys.stderr)
                console_errors.setLevel(max((console_level, logging.WARNING)))
                logger.addHandler(console_errors)

    if log_filename is not None:
        if not isinstance(log_filename, Path):
            log_filename = Path(log_filename)
        log_filename = log_filename.resolve().expanduser()
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(file_level)
        for existing_file_handler in [
            handler for handler in logger.handlers if type(handler) is logging.FileHandler
        ]:
            logger.removeHandler(existing_file_handler)
        logger.addHandler(file_handler)

    if log_format is None:
        log_format = '%(asctime)s | %(levelname)-8s | %(message)s'
    log_formatter = logging.Formatter(log_format)
    for handler in logger.handlers:
        handler.setFormatter(log_formatter)

    return logger



from datetime import date, datetime
from functools import partial
from getpass import getpass
from logging import Logger
from sqlite3 import Cursor
from typing import Any, Collection, Dict, List, Mapping, Sequence, Union, get_args as typing_get_args

import psycopg2
from psycopg2._psycopg import connection
from pyproj import CRS
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry, GEOMETRY_TYPES
from sshtunnel import SSHTunnelForwarder
from tablecrow.tables.base import DatabaseTable, parse_record_values, random_open_tcp_port

from ..utilities import guard_generic_alias, parse_hostname, split_hostname_port

SSH_DEFAULT_PORT = 22


class PostGresTable(DatabaseTable):
    DEFAULT_PORT = 5432
    FIELD_TYPES = {
        'NoneType': 'NULL',
        'bool': 'BOOL',
        'float': 'REAL',
        'int': 'INTEGER',
        'str': 'VARCHAR',
        'bytes': 'BYTEA',
        'date': 'DATE',
        'time': 'TIME',
        'datetime': 'TIMESTAMP',
        'timedelta': 'INTERVAL',
        'dict': 'HSTORE',
        'ipaddress': 'INET',
        **{geometry_type: 'GEOMETRY' for geometry_type in GEOMETRY_TYPES},
    }

    def __init__(
        self,
        hostname: str,
        table_name: str,
        database: str = None,
        fields: Dict[str, type] = None,
        primary_key: Union[str, List[str]] = None,
        crs: CRS = None,
        username: str = None,
        users: List[str] = None,
        logger: Logger = None,
        **kwargs,
    ):
        self.tunnel_credentials = {}

        if 'ssh_hostname' in kwargs and kwargs['ssh_hostname'] is not None:
            credentials = parse_hostname(kwargs['ssh_hostname'])
            ssh_hostname = credentials['hostname']
            ssh_port = credentials['port']
            if ssh_port is None:
                ssh_port = SSH_DEFAULT_PORT

            ssh_username = kwargs['ssh_username'] if 'ssh_username' in kwargs else None
            ssh_password = kwargs['ssh_password'] if 'ssh_password' in kwargs else None

            self.tunnel_credentials['ssh_hostname'] = ssh_hostname
            self.tunnel_credentials['ssh_port'] = ssh_port
            self.tunnel_credentials['ssh_username'] = ssh_username
            self.tunnel_credentials['ssh_password'] = ssh_password

        password = None
        if username is not None and ':' in username:
            username, password = username.split(':', 1)
        if 'password' in kwargs:
            password = kwargs['password']
        if password is None:
            password = getpass()
        self._DatabaseTable__password = password

        if database is None:
            database = 'postgres'

        super().__init__(
            resource=hostname,
            table_name=table_name,
            database=database,
            fields=fields,
            primary_key=primary_key,
            crs=crs,
            username=username,
            password=password,
            users=users,
            logger=logger,
        )

        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
            )

        if self.fields is None:
            with self.connection as connection:
                with connection.cursor() as cursor:
                    self._DatabaseTable__fields = database_table_fields(cursor, self.name)
            connection.close()

            if self.primary_key is None:
                self._DatabaseTable__primary_key = list(self.fields)[0]

        with self.connection as connection:
            with connection.cursor() as cursor:
                if database_has_table(cursor, self.name):
                    if database_table_is_inherited(cursor, self.name):
                        raise RuntimeError(
                            f'inheritance of table "{self.database}/{self.name}" will cause unexpected behaviour; aborting'
                        )

                    remote_fields = self.remote_fields
                    if list(remote_fields) != list(self.fields):
                        self.logger.warning(
                            f'schema of existing table "{self.database}/{self.name}" differs from given fields'
                        )

                        remote_fields_not_in_local_table = {
                            field: value
                            for field, value in remote_fields.items()
                            if field not in self.fields
                        }
                        if len(remote_fields_not_in_local_table) > 0:
                            self.logger.warning(
                                f'remote table has {len(remote_fields_not_in_local_table)} fields not in local table: {list(remote_fields_not_in_local_table)}'
                            )
                            self.logger.warning(
                                f'adding {len(remote_fields_not_in_local_table)} fields to local table: {list(remote_fields_not_in_local_table)}'
                            )

                            self._DatabaseTable__fields.update(
                                remote_fields_not_in_local_table
                            )
                            self._DatabaseTable__fields = {
                                field: self._DatabaseTable__fields[field]
                                for field in remote_fields
                            }

                        local_fields_not_in_remote_table = {
                            field: value
                            for field, value in self.fields.items()
                            if field not in remote_fields
                        }
                        if len(local_fields_not_in_remote_table) > 0:
                            self.logger.warning(
                                f'local table has {len(local_fields_not_in_remote_table)} fields not in remote table: {list(local_fields_not_in_remote_table)}'
                            )
                            self.logger.warning(
                                f'adding {len(local_fields_not_in_remote_table)} fields to remote table: {list(local_fields_not_in_remote_table)}'
                            )

                        if list(remote_fields) != list(self.fields):
                            self.logger.warning(
                                f'altering schema of "{self.database}/{self.name}"'
                            )
                            self.logger.debug(self.remote_fields)
                            self.logger.debug(self.fields)

                            copy_table_name = f'old_{self.name}'

                            if database_has_table(cursor, copy_table_name):
                                cursor.execute(f'DROP TABLE {copy_table_name};')

                            cursor.execute(
                                f'ALTER TABLE {self.name} RENAME TO {copy_table_name};'
                            )

                            cursor.execute(f'CREATE TABLE {self.name} ({self.schema});')
                            for user in self.users:
                                cursor.execute(
                                    f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.name} TO {user};'
                                )

                            copy_table_fields = list(
                                database_table_fields(cursor, copy_table_name)
                            )

                            cursor.execute(
                                f'INSERT INTO {self.name} ({", ".join(copy_table_fields)}) SELECT {", ".join(copy_table_fields)} FROM {copy_table_name};'
                            )

                            cursor.execute(f'DROP TABLE {copy_table_name};')
                else:
                    self.logger.debug(f'creating remote table "{self.database}/{self.name}"')
                    cursor.execute(f'CREATE TABLE {self.name} ({self.schema});')

                    for user in self.users:
                        cursor.execute(
                            f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.name} TO {user};'
                        )
        connection.close()
        if 'password' in kwargs:
            kwargs['password'] = '*****'
        self.kwargs = kwargs

    @property
    def hostname(self) -> str:
        return self.resource

    @property
    def tunnel(self) -> SSHTunnelForwarder:
        if 'ssh_hostname' in self.tunnel_credentials:
            port = split_hostname_port(self.resource)[-1]
            if port is None:
                port = self.DEFAULT_PORT
            tunnel = SSHTunnelForwarder(
                (self.tunnel_credentials['ssh_hostname'], self.tunnel_credentials['ssh_port']),
                ssh_username=self.tunnel_credentials['ssh_username'],
                ssh_password=self.tunnel_credentials['ssh_password'],
                remote_bind_address=('localhost', port),
                local_bind_address=('localhost', random_open_tcp_port()),
            )
            try:
                tunnel.start()
            except Exception as error:
                raise ConnectionError(error)
        else:
            tunnel = None
        return tunnel

    @property
    def connection(self) -> connection:
        connector = partial(
            psycopg2.connect,
            database=self.database,
            user=self.username,
            password=self._DatabaseTable__password,
        )

        tunnel = self.tunnel
        if tunnel is not None:
            connection = connector(host=tunnel.local_bind_host, port=tunnel.local_bind_port)
        else:
            connection = connector(host=self.hostname, port=self.port)

        return connection

    @property
    def exists(self) -> bool:
        with self.connection as connection:
            with connection.cursor() as cursor:
                exists = database_has_table(cursor, self.name)
        connection.close()
        return exists

    @property
    def schema(self) -> str:
        """ PostGres schema string """

        schema = []
        for field, field_type in self.fields.items():
            field_type = guard_generic_alias(field_type)

            if field_type in [list, tuple, Sequence, Collection]:
                field_type = [typing_get_args(field_type[0])]
            dimensions = 0
            while isinstance(field_type, Sequence) and not isinstance(field_type, str):
                if len(field_type) > 0:
                    field_type = field_type[0]
                else:
                    field_type = list
                dimensions += 1
            if isinstance(field_type, Mapping):
                field_type = dict

            try:
                field_type = self.FIELD_TYPES[field_type.__name__]
            except KeyError:
                raise TypeError(f'PostGres does not support type "{field_type}"')

            schema.append(f'{field} {field_type}{"[]" * dimensions}')

        schema.append(f'PRIMARY KEY({", ".join(self.primary_key)})')

        return ', '.join(schema)

    @property
    def remote_fields(self) -> Dict[str, type]:
        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
            )

        fields = None
        with self.connection as connection:
            with connection.cursor() as cursor:
                if database_has_table(cursor, self.name):
                    fields = database_table_fields(cursor, self.name)

                    for field, field_type in fields.items():
                        dimensions = field_type.count('_')
                        field_type = field_type.strip('_')

                        field_type = field_type.lower()
                        if field_type == 'geometry':
                            if field in self.fields:
                                fields[field] = self.fields[field]
                                continue

                        for python_type, postgres_type in self.FIELD_TYPES.items():
                            if postgres_type.lower() == field_type:
                                if field_type == 'geometry':
                                    if python_type not in globals():
                                        exec(f'from shapely.geometry import {python_type}')
                                field_type = eval(python_type)
                                break
                        else:
                            for python_type, postgres_type in self.FIELD_TYPES.items():
                                if python_type.lower() in field_type:
                                    field_type = eval(python_type)
                                    break
                            else:
                                field_type = str

                        for _ in range(dimensions):
                            field_type = [field_type]
                        fields[field] = field_type
                else:
                    fields = None
        connection.close()

        return fields

    @property
    def connected(self) -> bool:
        with self.connection as connection:
            try:
                with connection.cursor() as cursor:
                    cursor.execute('SELECT 1;')
                    cursor.fetchone()
                connected = True
            except:
                connected = False
        connection.close()
        return connected

    def records_where(
        self, where: Union[Mapping[str, Any], str, List[str]]
    ) -> List[Dict[str, Any]]:
        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
            )

        where_clause, where_values = self.__where_clause(where)

        with self.connection as connection:
            with connection.cursor() as cursor:
                if where_clause is None:
                    cursor.execute(f'SELECT {", ".join(self.fields.keys())} FROM {self.name};')
                else:
                    try:
                        cursor.execute(
                            f'SELECT * FROM {self.name} WHERE {where_clause};', where_values
                        )
                    except psycopg2.errors.UndefinedColumn as error:
                        raise KeyError(error)
                    except psycopg2.errors.SyntaxError as error:
                        raise SyntaxError(f'invalid SQL syntax - {error}')
                matching_records = cursor.fetchall()
        connection.close()

        matching_records = [
            parse_record_values(dict(zip(self.fields.keys(), record)), self.fields)
            for record in matching_records
        ]

        return matching_records

    def records_intersecting(
        self, geometry: BaseGeometry, crs: CRS = None, geometry_fields: List[str] = None
    ) -> List[Dict[str, Any]]:
        if crs is None:
            crs = self.crs

        if crs.to_epsg() is None:
            raise NotImplementedError(f'no EPSG code found for CRS "{crs}"')

        if geometry_fields is None or len(geometry_fields) == 0:
            geometry_fields = list(self.geometry_fields)

        where_clause = []
        where_values = []
        for field in geometry_fields:
            where_values.extend([geometry.wkt, crs.to_epsg()])
            geometry_string = f'ST_GeomFromText(%s, %s)'
            if crs != self.crs:
                geometry_string = f'ST_Transform({geometry_string}, %s)'
                where_values.append(self.crs.to_epsg())
            where_clause.append(f'ST_Intersects({field}, {geometry_string})')
        where_clause = ' OR '.join(where_clause)

        with self.connection as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    f'SELECT * FROM {self.name} WHERE {where_clause};', where_values
                )
                records = cursor.fetchall()
        connection.close()

        return [
            parse_record_values(dict(zip(self.fields.keys(), record)), self.fields)
            for record in records
        ]

    def insert(self, records: List[Dict[str, Any]]):
        if isinstance(records, dict):
            records = [records]

        if not all(field in record for field in self.primary_key for record in records):
            raise KeyError(
                f'one or more records does not contain primary key(s) "{self.primary_key}"'
            )

        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
            )

        with self.connection as connection:
            with connection.cursor() as cursor:
                for record in records:
                    if len(self.primary_key) == 1:
                        primary_key_string = self.primary_key[0]
                        primary_key_value = record[self.primary_key[0]]
                    else:
                        primary_key_string = f'({", ".join(self.primary_key)})'
                        primary_key_value = tuple(
                            record[primary_key] for primary_key in self.primary_key
                        )

                    record_fields_not_in_local_table = [
                        field for field in record if field not in self.fields
                    ]
                    if len(record_fields_not_in_local_table) > 0:
                        self.logger.warning(
                            f'record has {len(record_fields_not_in_local_table)} fields not in the local table'
                            f' that will not be inserted: {record_fields_not_in_local_table}'
                        )

                    local_fields_in_record = [
                        field for field in self.fields if field in record
                    ]
                    geometry_fields = [
                        field for field in self.geometry_fields if field in record
                    ]

                    columns = [
                        field
                        for field in local_fields_in_record
                        if field not in geometry_fields
                    ]
                    values = [
                        record[field]
                        for field in local_fields_in_record
                        if field not in geometry_fields
                    ]

                    for index, value in enumerate(values):
                        if isinstance(value, Collection) and not isinstance(
                            value, (str, list)
                        ):
                            values[index] = list(value)

                    if primary_key_value in self:
                        record_without_primary_key = {
                            column: value
                            for column, value in zip(columns, values)
                            if column not in self.primary_key
                        }
                        if len(record_without_primary_key) > 0:
                            if len(record_without_primary_key) > 1:
                                cursor.execute(
                                    f'UPDATE {self.name} SET ({", ".join(record_without_primary_key.keys())}) = %s '
                                    f'WHERE {primary_key_string} = %s;',
                                    [
                                        tuple(record_without_primary_key.values()),
                                        primary_key_value,
                                    ],
                                )
                            else:
                                cursor.execute(
                                    f'UPDATE {self.name} SET {tuple(record_without_primary_key.keys())[0]} = %s '
                                    f'WHERE {primary_key_string} = %s;',
                                    [
                                        tuple(record_without_primary_key.values())[0],
                                        primary_key_value,
                                    ],
                                )
                    else:
                        cursor.execute(
                            f'INSERT INTO {self.name} ({", ".join(columns)}) VALUES %s;',
                            [tuple(values)],
                        )

                    if len(geometry_fields) > 0:
                        geometries = {
                            field: record[field]
                            for field in geometry_fields
                            if record[field] is not None
                        }

                        for field, geometry in geometries.items():
                            cursor.execute(
                                f'UPDATE {self.name} SET {field} = ST_GeomFromText(%s, %s) '
                                f'WHERE {primary_key_string} = %s;',
                                [geometry.wkt, self.crs.to_epsg(), primary_key_value],
                            )
        connection.close()

    def delete_where(self, where: Union[Mapping[str, Any], str, List[str]]):
        if not self.connected:
            raise ConnectionError(
                f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
            )

        where_clause, where_values = self.__where_clause(where)

        with self.connection as connection:
            with connection.cursor() as cursor:
                if where_clause is None:
                    cursor.execute(f'TRUNCATE {self.name};')
                else:
                    try:
                        cursor.execute(
                            f'DELETE FROM {self.name} WHERE {where_clause};', where_values
                        )
                    except psycopg2.errors.UndefinedColumn as error:
                        raise KeyError(error)
                    except psycopg2.errors.SyntaxError as error:
                        raise SyntaxError(f'invalid SQL syntax - {error}')
        connection.close()

    def __len__(self) -> int:
        with self.connection as connection:
            with connection.cursor() as cursor:
                cursor.execute(f'SELECT COUNT(*) FROM {self.name};')
                length = cursor.fetchone()[0]
        connection.close()
        return length

    def delete_table(self):
        with self.connection as connection:
            with connection.cursor() as cursor:
                cursor.execute(f'DROP TABLE {self.name};')
        connection.close()

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}({repr(self.database)}, {repr(self.name)}, {repr(self.fields)}, {repr(self.primary_key)}, '
            f'{repr(self.hostname)}, {repr(self.crs.to_epsg()) if self.crs is not None else None}, '
            f'{repr(self.username)}, {repr(self.users)}'
            f'{", " if len(self.kwargs) > 0 else ""}'
            f'{", ".join(key + "=" + repr(value) for key, value in self.kwargs.items())})'
        )

    def __where_clause(self, where: Dict[str, Union[Any, List]]) -> (str, List):
        if (
            where is not None
            and not isinstance(where, Sequence)
            and not isinstance(where, dict)
        ):
            raise NotImplementedError(f'unsupported query type "{type(where)}"')

        if where is None or len(where) == 0:
            where_clause = None
            where_values = None
        else:
            fields = None
            where_values = []
            if isinstance(where, str):
                where_clause = where
            elif isinstance(where, dict):
                where_clause = []
                for field, value in where.items():
                    field_type = self.fields[field]
                    if isinstance(value, BaseGeometry) or isinstance(
                        value, BaseMultipartGeometry
                    ):
                        where_clause.append(f'{field} = ST_GeomFromText(%s, %s)')
                        where_values.extend([value.wkt, self.crs.to_epsg()])
                    else:
                        if isinstance(field_type, list):
                            if not isinstance(value, Sequence) or isinstance(value, str):
                                statement = f'%s = ANY({field})'
                            else:
                                if fields is None:
                                    with self.connection as connection:
                                        with connection.cursor() as cursor:
                                            fields = database_table_fields(cursor, self.name)
                                    connection.close()
                                field_type = fields[field]
                                dimensions = field_type.count('_')
                                field_type = field_type.strip('_')
                                statement = f'{field} = %s::{field_type}{"[]" * dimensions}'
                        elif value is None:
                            statement = f'{field} IS %s'
                        elif isinstance(value, Sequence) and not isinstance(value, str):
                            statement = f'{field} IN %s'
                            value = tuple(value)
                        elif isinstance(value, str) and '%' in value:
                            statement = f'{field} ILIKE %s'
                        else:
                            if isinstance(value, datetime):
                                value = f'{value:%Y%m%d %H%M%S}'
                            elif isinstance(value, date):
                                value = f'{value:%Y%m%d}'
                            statement = f'{field} = %s'
                        where_values.append(value)
                        where_clause.append(statement)
                where_clause = ' AND '.join(where_clause)
            else:
                where_clause = ' AND '.join(where)

            if len(where_values) == 0:
                where_values = None

        return where_clause, where_values


def database_tables(cursor: Cursor, user_defined: bool = True) -> List[str]:
    """
    list of tables within the given PostGreSQL database

    :param cursor: psycopg2 cursor
    :return: list of table names
    """

    # query = "SELECT table_name FROM information_schema.tables WHERE table_schema='public'"
    query = f"SELECT relname FROM pg_class WHERE relkind='r'"
    if user_defined:
        query += " AND relname !~ '^(pg_|sql_)'"
    query += ';'

    cursor.execute(query)
    return [record[0] for record in cursor.fetchall()]


def database_has_table(cursor: psycopg2._psycopg.cursor, table: str) -> bool:
    """
    whether the given table exists within the given PostGreSQL database

    :param cursor: psycopg2 cursor
    :param table: name of table
    :return: whether table exists
    """

    cursor.execute(
        # f'SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE table_name=%s);',
        f'SELECT EXISTS(SELECT 1 FROM pg_class WHERE relname=%s);',
        [table.lower()],
    )
    return cursor.fetchone()[0]


def database_table_is_inherited(cursor: psycopg2._psycopg.cursor, table: str) -> bool:
    """
    whether the given PostGreSQL table is inherited

    :param cursor: psycopg2 cursor
    :param table: name of table
    :return: whether table is inherited
    """

    cursor.execute(
        f'SELECT EXISTS(SELECT 1 FROM pg_catalog.pg_inherits WHERE inhrelid=%s::regclass);',
        [f'public.{table}'],
    )
    return cursor.fetchone()[0]


def database_table_fields(cursor: psycopg2._psycopg.cursor, table: str) -> Dict[str, str]:
    """
    field names and data types of the given table, within the given PostGreSQL database

    :param cursor: psycopg2 cursor
    :param table: name of table
    :return: mapping of column names to the PostGres data type
    """

    cursor.execute(
        f'SELECT column_name, udt_name FROM information_schema.columns WHERE table_name=%s;',
        [table],
    )
    return {record[0]: record[1] for record in cursor.fetchall()}



#!/usr/bin/env python
from setuptools import config, find_packages, setup

try:
    from dunamai import Version
except ImportError:
    import subprocess
    import sys

    subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'dunamai'])
    from dunamai import Version

metadata = config.read_configuration('setup.cfg')['metadata']

setup(
    name=metadata['name'],
    version=Version.from_any_vcs().serialize(),
    author=metadata['author'],
    author_email=metadata['author_email'],
    description=metadata['description'],
    long_description=metadata['long_description'],
    long_description_content_type='text/markdown',
    url=metadata['url'],
    packages=find_packages(),
    python_requires='>=3.8',
    setup_requires=['dunamai', 'setuptools>=41.2'],
    install_requires=[
        'psycopg2-binary',
        'pyproj',
        'python-dateutil',
        'shapely',
        'spatialite',
        'sshtunnel',
    ],
    extras_require={
        'testing': ['pytest', 'pytest-cov', 'pytest-xdist'],
        'development': ['dunamai', 'flake8', 'isort', 'oitnb'],
        'documentation': ['dunamai', 'm2r2', 'sphinx', 'sphinx-rtd-theme'],
    },
)
